/************************************************************************/
/*  Copyright (c) 1997-2000 Nalante, Inc., 245 Falconer End, Edmonton,  */
/*  Alberta, Canada, T6R 2V6, Phone: (780) 432-5288, Fax: (780) 988-0184*/
/*  All Rights Reserved                                                 */
/*                                                                      */
/*  THIS IS UNPUBLISHED PROPRIETARY SOURCE CODE OF Nalante, Inc.        */
/*  The copyright notice above does not evidence any                    */
/*  actual or intended publication of such source code.                 */
/*                                                                      */
/*  Unauthorized copy or distribution of this file is prohibited.       */
/************************************************************************/


#ifndef PTREE_H
#define PTREE_H

#ifndef TRSTREE_H
#include "trstree.h"
#endif

#ifndef TARRAY_H
#include "tarray.h"
#endif

class ParseNode;
class StrBufs;
class TNode;

/// Feature-Value pairs. 
struct FeatureValue {
  const char* _feature;
  const char* _value;
  FeatureValue(const char* f = 0, const char* v = 0) {
    _feature = f;
    _value = v;
  }
};

//typedef enum { CONSTITUENCY, DEPENDENCY } ParseType;

/**
   The data structure for storing parse trees.
 */
class CLASS_DECLSPEC CLASS_DECLSPEC ParseTree {
protected:
//  /// the type of the tree: CONSTITUENCY or DEPENDENCY
//  ParseType _type;

  /// the root node of the tree
  ParseNode* _root;

  /// buffer for the strings used in the tree (no linguistic significance).
  StrBufs* _buf;    
public:
  ~ParseTree();

//  ParseTree(ParseType type);
  ParseTree();

//  /// return the type of the tree: CONSTITUENCY or DEPENDENCY
//  ParseType type() const { return _type;}

  /// return the root node of the tree
  const ParseNode* root() const { return _root;}

  /// set the root node of the tree
  void root(ParseNode* root) { _root = root;}

  /// print the tree on a output stream
  void print(ostream& ostrm);

  /// make a copy of the concatenation of s1, s2, ...
  const char* get_str(
    const char* s1, const char* s2=0, const char* s3=0, 
    const char* s4=0, const char* s5=0
    );

  /// remove the current tree and get ready for another parse tree
  void reset();

  /// Create a ParseNode from a node in a packed parse forest (TNode*)
  ParseNode* get_pnode(TNode* tn);
};

/**
   A node in a parse tree.
 */
class CLASS_DECLSPEC CLASS_DECLSPEC ParseNode : public TRSTree<ParseNode*> {
protected:
  /// the parse tree to which this node belongs
  ParseTree* _tree;

  /// the surface words that this node represents. For non-terminal
  /// nodes in a constituency tree or empty categories (in both
  /// constituency tree and dependency tree, _word is 0;
  const char* _word;

  /// The root form of _word;
  const char* _root;

  /// The grammatical category this node represents. For dependency
  /// trees, _category is always the part of speech of the _word
  const char* _category;

  /// The grammatical relationship between this node and its parent
  const char* _relation;

  /// The label assigned to a node in DEPENDENCY format
  const char* _label;

  /// The list of grammatical features that this word possesses
  TArray<FeatureValue> _features;

  /// The head child of this node (only used in constituency trees
  ParseNode* _head_child;

  /// The antecedent of an empty category (in the same parse tree).
  ParseNode* _antecedent;

  /// the starting word index of the category represented by the node.
  /// For constituency trees, this is the lowest word index in the subtree
  /// that is underneath this node. For dependency trees, this is the 
  /// starting word index of the head word.
  int _low;
  
  /// the ending word index of the category represented by the node.
  /// For constituency trees, this is the highest word index in the subtree
  /// that is underneath this node. For dependency trees, this is the 
  /// ending word index of the head word.
  int _high;

  /// the word index of the head word.
  int _head_pos;

  /// the total weight of a subtree, used only in a constituency tree
  float _weight;
public:
  ~ParseNode();

  const ParseTree* tree() const { return _tree;}

  const char* word() const { return _word;}
  void word(const char* word) { _word = _tree->get_str(word);}

  const char* root() const { return _root;}
  void root(const char* root) { _root = _tree->get_str(root);}

  const char* category() const { return _category;}
  void category(const char* category) { _category = _tree->get_str(category);}

  const char* relation() const { return _relation;}
  void relation(const char* relation) { _relation = _tree->get_str(relation);}

  const char* label() const { return _label;}
  void label(const char* label) { _label = _tree->get_str(label);}

  void add_weight(float w) { _weight += w;}
  float weight() const { return _weight;}

  int low() const { return _low;}
  void low(int low) { _low = low;}

  int high() const { return _high;}
  void high(int high) { _high = high;}

  int head_pos() const { return _head_pos;}
  void head_pos(int pos) { _head_pos = pos;}

  const ParseNode* head_child() const { return _head_child;}
  void head_child(ParseNode* head_child) { _head_child = head_child;}

  const ParseNode* antecedent() const { return _antecedent;}
  void antecedent(ParseNode* antecedent) { _antecedent = antecedent;}

  /// return the list of features
  TArray<FeatureValue>& features() { return _features;}

  /// insert the subtree as a child of this node according to the surface 
  /// word order
  void ord_insert(ParseNode* subtree);

  /// print the information pertaining to this node
  ostream& print_root(ostream& strm) const;

  ParseNode(ParseTree& tree);
};

/** This function is used to initialize the parser.  It must be called
    before before the parser is used. */
CLASS_DECLSPEC extern void initialize_minipar(const char* paths);

/** Define the following features to be extracted and stored in the
    ParseTrees. The variable features is a list of feature names
    separated by spaces.*/
CLASS_DECLSPEC extern void extract_features(const char* features);

/** Parse a sentence that is stored in an array of characters. The
    sentence is tokenized automatically.  */
CLASS_DECLSPEC extern void parse(const char* sentence, ParseTree& parse_tree);

/** Parse a sentence that is represented by an array of character
    pointers. Each pointer points to a words. The end of sentence is
    marked by a null pointer.  */
CLASS_DECLSPEC extern void parse(const char* words[], ParseTree& parse_tree);

CLASS_DECLSPEC extern void interpret_command_line(const char* line);

/** print the contents of a lexicon entry for the given word in the
    ostream. */
CLASS_DECLSPEC extern void print_lexicon_entry(const char* word, ostream& ostrm, const char* lexiconName = "English-lexicon");
#endif /* PTREE_H */
