/************************************************************************/
/*  Copyright (c) 1997-2000 Nalante, Inc., 245 Falconer End, Edmonton,  */
/*  Alberta, Canada, T6R 2V6, Phone: (780) 432-5288, Fax: (780) 988-0184*/
/*  All Rights Reserved                                                 */
/*                                                                      */
/*  THIS IS UNPUBLISHED PROPRIETARY SOURCE CODE OF Nalante, Inc.        */
/*  The copyright notice above does not evidence any                    */
/*  actual or intended publication of such source code.                 */
/*                                                                      */
/*  Unauthorized copy or distribution of this file is prohibited.       */
/************************************************************************/



#ifndef RSLIST_H
#define RSLIST_H

#ifndef TSLIST_H
#include "tslist.h"
#endif

////                    CLASS RSlist
////    Recursive Slist. An RSlist is an Slist whose elements
////    are also RSlist's. A tree structure can easily be
////    represented by an RSlist.

class RSlist;

class CLASS_DECLSPEC RSlist : public Slist {
  /* Recursive Slist */
friend class RSlistIterator;
friend class PreIterator;
friend class BreadthIterator;
friend class PostIterator;
public:
  int atom() const { return empty();}
  void append_list(RSlist& list) {Slist::append_list(list);}
  RSlist* get() {return (RSlist*) Slist::get();}
  RSlist* get(int i) const { return (RSlist*) Slist::get(i);}
  RSlist* replace(int i, RSlist* l) {
    return (RSlist*) Slist::replace(i, l);
  }
  RSlist* del(RSlist* l) { return (RSlist*) Slist::del(l);}
  RSlist* search(RSlist* l) { return (RSlist*) Slist::search(l);}
  RSlist* search(const Ent arg, TwoArgPred pred, int neg=0) const {
    return (RSlist*) Slist::search(arg, pred, neg);
  }
  RSlist* top() const { return (RSlist*) Slist::top();}
  RSlist* first() const { return (RSlist*) Slist::first();}
  RSlist* last() { return (RSlist*) Slist::last();}
  RSlist* reduce(RSlist* seed, TwoArgFunc func) {
    return (RSlist*) Slist::reduce(seed, func);
  }
  RSlist() {}
  void print(const RSlist* (*printer)(const RSlist*, ostream&), ostream& ostrm, int depth = 0) const;
};

/*
    Suppose an RSlist is as follows:

               a
              /  \
             b    c
            / \    \
           d   e    f

    The iterators will visit the nodes in the following
    sequences:
    1. PreIterator:
       a b d e c f
    2. PostIterator:
       d e b f c a
    3. BreadthIterator:
       a b c d e f



        CLASS RSlistIterator
    Iterator of RSlists.

    Internal Data:
    Slink* ce;
        points to the Slink whose e field points to the
        next RSlist to be visited.
    RSlist* cs;
        points to the RSlist to which the RSlist pointed
        to be ce belongs.
    Slist ce_stack;
        The stack of the ce's of the parents of the current
        list being visited.
    RSlist cs_stack;
        The stack of parents fo the current list.

    Interface:
    int finished():
        Return true at the end of traversal.
    RSlist* parent():
        Return the parent list of the current list
        being visited.
    RSlistIterator():
        Constructor.

    The subclass of RSlistIterator implements methods
    next() and skip(). The former move the current
    element to the next one. The latter skip the
    current subtree.

    delete_current( iterator): MACRO will expand to
    del_current():
        delete the current RSlist and all its subtrees
        it return the delted tree; it then set current
        to the next RSlist (with deleted field = 1)

    delete_parent( iterator, level): MACRO will expand to
    del_parent(level):
        delete the parent of current RSlist. level =0 (default)
        delete its direct parent; level =1, delete its grandparent; ...
        it return the deleted parent tree; it then set current
        to the next RSlist (with deleted field = 1)

    insert_before_current( iterator, tree): MACRO will expand to
    interator::insert_before_current(tree):
        Insert the tree before the current RSlist. The newly inserted
        RSlist will NOT be traversed during the current FORALL loop

    insert_after_current( iterator, tree): MACRO will expand to
    interator::insert_after_current(tree):
        Insert the tree after the current RSlist. The newly inserted
        RSlist will BE traversed during the current FORALL loop
*/

class CLASS_DECLSPEC RSlistIterator : public Iterator {
protected:
  Slink _dummy;
  RSlist* _above_top;
  Slist ce_stack;
  Slist cs_stack;
  Slink* ce;
  RSlist* cs;
  int _depth;

  void push(RSlist* list, Slink* link) {
    cs_stack.insert(list);
    ce_stack.insert(link);
    _depth++;
  }
  void pop() {
    cs = (RSlist*) cs_stack.get();
    ce = (Slink*) ce_stack.get();
    _depth--;
  }
  virtual RSlist* v_skip() = 0;
  virtual Ent v_del_current() = 0;
  virtual Ent v_del_parent(int level) = 0;
public:
  int depth() { return _depth;}
  int end_of_sublist() { return ce==cs->last_slink();}
  int finished() { return ce == 0; }
  RSlist* parent() { return cs; }
  RSlist* skip() { return v_skip();}
  RSlist* current();
  void insert_before_current(RSlist* x);
  void insert_after_current(RSlist* x);
  Ent del_current() { return v_del_current();}
  Ent del_parent(int level=1) { return v_del_parent(level);}
  RSlistIterator(const RSlist* slist);
  virtual ~RSlistIterator();
};

class CLASS_DECLSPEC PreIterator : public RSlistIterator {
  // A node is visited before all of its children
  void reset(const RSlist* slist);
public:
  PreIterator(const RSlist* slist) : RSlistIterator(slist) {reset(_above_top);}
  RSlist* next();
  RSlist* v_skip(); /* skip visiting the elements of current list */
  void reset();
  Ent v_del_current();
  Ent v_del_parent(int level=1);
};

class CLASS_DECLSPEC PostIterator : public RSlistIterator {
  // A node is visited after all its children
  void reset(const RSlist* slist);
public:
  PostIterator(const RSlist* slist) : RSlistIterator(slist) {
    reset(_above_top);
  }
  RSlist* next();
  RSlist* v_skip(); /* skip the rest of the list */
  void reset();
  Ent v_del_current();
  Ent v_del_parent(int level=1);
};

class CLASS_DECLSPEC BreadthIterator : public RSlistIterator {
  /* elements of the list is visitied in breadth first order */
protected:
  Slink* level_boundary;

  /* note that the cs_stack, ce_stack are used as
     queues instead of stacks */
  void push() { }
  void pop() {
    cs = (RSlist*) cs_stack.get();
    ce = (Slink*) ce_stack.get();
  }
  void reset(const RSlist* slist);
public:
  RSlist* next();
  RSlist* skip();
  void reset();
  BreadthIterator(const RSlist* slist) 
    : RSlistIterator(slist) { reset(_above_top);}
  Ent v_del_current();
};

#endif /* RSLIST_H */
