/************************************************************************/
/*  Copyright (c) 1997-2000 Nalante, Inc., 245 Falconer End, Edmonton,  */
/*  Alberta, Canada, T6R 2V6, Phone: (780) 432-5288, Fax: (780) 988-0184*/
/*  All Rights Reserved                                                 */
/*                                                                      */
/*  THIS IS UNPUBLISHED PROPRIETARY SOURCE CODE OF Nalante, Inc.        */
/*  The copyright notice above does not evidence any                    */
/*  actual or intended publication of such source code.                 */
/*                                                                      */
/*  Unauthorized copy or distribution of this file is prohibited.       */
/************************************************************************/



#ifndef SLIST_H
#define SLIST_H

#include <cstddef>

#ifndef ENT_H
#include "ent.h"
#endif

#ifndef FORALL_H
#include "forall.h"
#endif


class SlistIterator;

class CLASS_DECLSPEC Slink {
private:
  enum { NALL = 250};
  static Slink* nfree;
  static void* add_chunks();
protected:
  Slink* _next;
  Ent _data;
public:
  Ent data() const { return _data;}
  Slink* next() const { return _next;}
  void next(Slink* n) { _next =  n;}
  void data(const Ent d) { _data = d;}
  Slink(Ent d, Slink* n=0) { _next = n; _data = d;}
  Slink() { _next = 0; _data = 0;}
  void* operator new(size_t) {
    register Slink* p = nfree;
    if (p==0)
      p = (Slink*) add_chunks();
    else nfree = p->next();
    return p;
  }

  void operator delete(void* p, size_t) {
    ((Slink*) p)->next(nfree);
    nfree = (Slink*) p;
  }
};

/**
    An Slist is a circular single linked list. Elements can be added
    and removed at the both ends of the list efficiently. The
    definition of Slist is based on the "slist" in Stroustrup's C++
    book.

    The elements in an Slist are linked by a sequence of Sinks. The
    each Slink contains a point to another Slink and a pointer to the
    element in the list. The last slink's next field points the first
    slink. An Slist object contains a point to the last slink. This is
    illustrated in the diagram bellow:

                          _last
                            |
                           \|/
     _____     _____     _____
    |__|__|-->|__|__|-->|__|__|---|
      |                           |
      |----<----------------------|

*/

class CLASS_DECLSPEC Slist {
  friend class SlistIterator;
  friend class SlistFollowingIterator;
protected:
  Slink* last_slink() const;
  Slink* first_slink() const;
public:
  /**
    ord_insert return a value if there is an element in the list that
    is equal to the element being inserted. Therefore, if no_dup==1
    and return value!=0, then the element is not inserted.  otherwise,
    the element is inserted.
  */
  Ent ord_insert(const Ent a, TwoArgPred cmp, int no_dup = 0) {
      //insert into an ordered Slist;
    Slink* place;
    Ent v = ord_search(a, cmp, place);
    if (no_dup==0 || v==0) 
      insert_at(a, place);
    return v;
  }

  /**
    ord_search returns the element in the list that is equal to the
    element being searched. it returns 0 if the list does not contain
    such an element. 
  */
  Ent ord_search(const Ent a, TwoArgPred cmp, Slink*& place);


  /// delete an element from an ordered Slist;
  Ent ord_del(const Ent a, TwoArgPred cmp);

  /// sort the Slist according to the ordering defined by cmp.
  void sort(TwoArgPred cmp);

  /// insert element a at place, the value of place is usually obtained from ord_search
  void insert_at(Ent a, Slink* place);

  /// insert an element at the beginning of the list
  void insert(const Ent a);

  /// append an element at the end of the list
  void append(const Ent a);     

  /// insert the element at the specified position
  void insert(int index, const Ent a);

  /// same as append
  void add(const Ent a);

  /// append a Slist, the argument list becomes empty;  
  void append_list(Slist&); 

  /// prepend a Slist, the argument list becomes empty;
  void prepend_list(Slist&); 

  /// remove the first element from the list and return it
  Ent get(); 

  /// return the ith element; do not remove it
  Ent get(int) const;
  
  /// replace the ith element from the list
  Ent replace(int, const Ent);

  /// delete an element from the list
  Ent del(const Ent a); 

  /// search the list for an element
  Ent search(const Ent a) const;

  /// search an element x in the list such that pred(x) is true (or 0 if neg!=0)
  Ent search(OneArgPred pred, int neg = 0) const;

  /// search an element x in the list such that pred(x, a) is true (or 0 if neg!=0)
  Ent search(const Ent a, TwoArgPred pred, int neg = 0) const;

  /// find the minimum (or maximum if neg!=0) element in the list
  Ent find_minimum(TwoArgPred cmp, int neg = 0) const;

  /** remove all elements from the list (the memory occupied by these elements are 
      not deallocated
  */
  void clear(); // remove all links

  /// return true if the list empty
  int empty() const;

  /// return number of elements in Slist
  int length() const; 

  /// return the first element in the list
  Ent first() const;

  /// return the last element in the list
  Ent last() const;

  /// make a copy of the list l
  void copy(const Slist& l);

  Ent reduce(const Ent seed, TwoArgFunc func);
  void reverse();
  void map(Slist& a, OneArgFunc f);
  void combine(Slist& a, Slist& b, TwoArgFunc f);
  void remove(Ent arg, TwoArgPred pred, Slist& removed);
  void select(Slist& b, OneArgPred pred, int neg = 0);
  void select(Slist& b, Ent arg, TwoArgPred pred, int neg = 0);
  void collect(Slist& b, OneArgFunc func);
  void collect(Slist& b, Ent arg, TwoArgFunc func);
  void apply(OneArgProc proc);
  void apply(Ent arg, TwoArgProc proc);
  void merge(Slist& a, Slist& b, TwoArgPred cmp);
  Slist();
  Slist(const Slist& list) { _last = 0; copy(list);}
  Slist& operator = (const Slist& list) { copy(list); return *this;}
  ~Slist();
  Ent pop();
  void push(Ent p);
  Ent top() const;
protected:
  Slink* _last; //_last->_next is head of list
  Slink* preceding(Slink* ce);
};


//                      CLASS SlistIterator
//    An SlistIterator is a iterator for an Slist. The
//    basic idea also comes from Stroustrup's C++ book. The
//    loops for enumerating elements in an Slist are as
//    follows:
//
//      SlistIterator it(&aSlist);
//
//      for (; !it.finished(); it.next()) {
//        use(it.current());
//      }
//
//    or
//      
//      while (!it.finished()) {
//        if (pred(it.current()))
//          it.next();
//        else
//          it.del_current();
//      }
//
//    Note that del_current delete the current element from
//    the list and move the current element to the next
//    element. The method
//
//      insert_before_current(Ent element)
//
//    inserts the element before the current element. The
//    current element remain unchanged.
//
//    The method operator() is implemented for
//    compatibility reasons. Its use is
//    discouraged. Instead, one should use finished() and
//    next() to iterate through the list. Such iteration is
//    in the same style as traversers for RSlist.


class CLASS_DECLSPEC SlistIterator : public Iterator {
protected:
  Slink* ce; // the Slink of the current element;
  Slist* cs; // current Slist;
  friend class SlistFollowingIterator;
public:
  SlistIterator(const Slist* s);
  SlistIterator(const Slist& s);
  Ent next();
  int finished() const;
  int at_start() const;
  void reset(const Slist* s = 0);
  Ent current() const;
  void insert_before_current(Ent x);
  void insert_after_current(Ent x);
  Ent del_current(); //delete the current element;
};

class CLASS_DECLSPEC SlistFollowingIterator : public SlistIterator {
protected:
  SlistIterator* _followed;
public:
  Ent del_current(); //delete the current element;
  SlistFollowingIterator(SlistIterator& it);
};

inline Slink* Slist::last_slink() const 
{
  return _last;
}

inline Slink* Slist::first_slink() const 
{
  return (Slink*) _last->next();
}

inline void Slist::add(const Ent a) 
{
  append(a); 
} 

inline int Slist::empty() const 
{
  return _last==0; // true if empty
}

inline Ent Slist::first() const 
{
  return (_last==0)?0:first_slink()->data();
}

inline Ent Slist::last() const 
{
  return (_last==0)?0:_last->data();
}

inline Slist::Slist()   
{
  _last=0; 
}

inline Ent Slist::pop() 
{
  return get();
}

inline void Slist::push(Ent p) 
{
  insert(p);
}

inline Ent Slist::top() const 
{
  return first();
}

inline Ent SlistIterator::current() const
{ 
  if (ce) return ce->next()->data();
  else return 0;
}

inline Ent SlistIterator::next() 
{
  // return  to indicate end of iteration
  // not ideal for all types; fine for pointers
  if (ce) {
    ce = ce->next();
    if (cs && ce == cs->last_slink()) ce = 0;
    return current();
  }
  else return 0;
}

inline int SlistIterator::finished() const
{
 return ce==0; 
}

inline int SlistIterator::at_start() const
{
 return ce==cs->last_slink();  // true if at the start
}

inline void SlistIterator::reset(const Slist* s) 
{ 
  if (s) cs = (Slist*) s; 
  if (cs) ce = cs->last_slink(); 
  else ce = 0;
  updated(0);
}



#endif /* SLIST_H */
