/************************************************************************/
/*  Copyright (c) 1997-2000 Nalante, Inc., 245 Falconer End, Edmonton,  */
/*  Alberta, Canada, T6R 2V6, Phone: (780) 432-5288, Fax: (780) 988-0184*/
/*  All Rights Reserved                                                 */
/*                                                                      */
/*  THIS IS UNPUBLISHED PROPRIETARY SOURCE CODE OF Nalante, Inc.        */
/*  The copyright notice above does not evidence any                    */
/*  actual or intended publication of such source code.                 */
/*                                                                      */
/*  Unauthorized copy or distribution of this file is prohibited.       */
/************************************************************************/



#ifndef TARRAY_H
#define TARRAY_H

#include <iostream>
#include <cassert>
#include <cstdlib>
#include <cstring>

#ifndef FORALL_H
#include "forall.h"
#endif

/**
   An expandable array. Elements in the array can be accessed with the
   [] operator, e.g., a[i]. If more elements are added to the array
   than its current capacity, the array will grow automatically, the
   previous elements in the array are copied to the newly allocated
   spaces. The data type stored in a TArray must have two constructors
   defined:
     Type() and Type(const Type&)

   The first one is used in the initialization of an array. The second
   one is used in copying elements.
 */
template<class CLASS_DECLSPEC Type> class TArray { 
protected:
  /// the size of _data array.
  int   _size; 

  /// one index higher than the last element in the list;
  int   _high; 
  
  /// the array that holds the data.
  Type* _data; 

protected:

  /// what to do when the index is out of bound
  void out_of_bounds(int index) const {
    cerr << "Array index out of bounds: " << index 
         << " It should be between 0 and " << _high-1 << endl;
    assert(0);
  }

  /// the expand the array capacity when the current value is not enough
  void grow(int len = -1) {
    if (len <= 0) 
      len = grow_size();
    Type* d = new Type[_size + len];
    if (_data) {
      int i;
      for (i = 0; i<length(); i++)
        d[i] = _data[i];
      delete [] _data;
    }
    _size += len;
    _data = d;
  }

  /// decide how much to grow
  int grow_size() { return (size()>>2) + 4;}

  /// reduce the array capacity
  void shrink() {
    if (size() > 20 && length() < (size() >> 1)) {
      Type* d = new Type[length()];
      int i;
      for (i = 0; i<length(); i++)
        d[++i] = _data[i];
      delete[] _data;
      _data = d;
      _size = length();
    }
  }

  int bi_search_aux(int l, int h, Type* key, int& index, int (*pred)(Type*, Type*))  {
    int cmp;
    if (h-l<10) {
      for (int i = l; i<=h; i++) {
        cmp = (*pred)(_data+i, key);
        if (cmp==0) { index = i; return 1;}
        else if (cmp>0) { index = i; return 0;}
      }
      index = h+1;
      return 0;
    }
    int mid = (l+h)>>1;
    cmp = (*pred)(_data+mid, key);
    if (cmp>0)
      return bi_search_aux(l, mid-1, key, index, pred);
    else if (cmp<0)
      return bi_search_aux(mid+1, h, key, index, pred);
    index = mid;
    return 1;
  }

public:
  /// the current capacity of the array
  int size() const { return _size;}

  /// Constructor
  TArray() {
    _size = 0;
    _high = 0;
    _data = 0;
  }

  /// Construct an array with capacity==size
  TArray(int size) {
    _size = 0;
    _data = 0;
    clear(size);
  }

  /// Copy constructor
  TArray(const TArray<Type>& array) {
    _size = 0;
    _data = 0;
    copy(array);
  }

  /// Destructor
  ~TArray() {
    if (_data) delete[] _data;
  }

  /// fill the array up to index or to capacity if index=-1
  void fill(int index = -1) {
    if (index<0)
      _high = _size;
    else {
      if (index>_size)
        grow(index-_size+1);
      _high = index;
    }
  }

  /// Move all the elements in al to this array. al becomes empty.
  void transfer(TArray<Type>& al) {
    if (_data)
      delete [] _data;
    _size = al._size;
    _high = al._high;
    _data = al._data;
    al._data = 0;
    al._size = 0;
    al._high = 0;
  }

  /// Copy all the elements in al to this, al remains unchanged
  void copy(const TArray<Type>& al) {
    if (al.length()>0) {
      Type* d = new Type[al.length()];
      memcpy(d, al._data, al.length()*sizeof(Type));
      init(d, al.length());
    }
    else 
      clear();
  }

  /**
     Create an empty place in the array (usually for assignment).
     Examples:
       array.get_vacant()  = element;
       array.get_vacant(i) = element;
     The first assignment statement puts the element at the end of the
     array. The second assignment statement puts the element at
     position i. The previous elements after position i are shifted by
     one position in the direction of higher index.
   */
  Type& get_vacant(int index = -1) {
    if (index<0)
      index = _high;
    if (full()) 
      grow();
    int i;
    for (i = _high; i>index; i--)
      _data[i] = _data[i-1];
    ++_high;
    return _data[index];
  }

  /// initialize an array with the initial elements in d
  void init(Type* d, int number) {
    if (_data) 
      delete[] _data;
    _high = _size = number;
    _data = d;
  }

  /// clear all elements in the array
  void clear() { _high = 0;}

  /// clear all elements in the array and set capacity to size. If size==-1, the current capacity is used.
  void clear(int size) {
    if (size>_size) {
      if (_data) delete [] _data;
      _data = new Type[size];
      _size = size;
    }
    else if (size==0) {
      delete[] _data;
      _data = 0;
    }
    _high = 0;
  }

  /// check whether an index is within the current range
  int valid(int index) const { return index >=0 && index <_high;}

  /// return true if the array is empty
  int empty() const  { return _high==0;}

  /// return the current number of documents in the array
  int length() const  { return _high;}

  /// return true if the array has reached full capacity.
  int full() const  {   return _size == _high || _data==0;}

  /// get the reference to an element in an array according to its index
  Type& operator [] (int index) const {
    if (index<0 || index>=_high)
      out_of_bounds(index);
    return _data[index];
  }

  /// return the reference to the last element in the array
  Type& last() const {
    if (_high<=0)
      out_of_bounds(_high-1);
    return _data[_high-1];
  }

  /// remove all the elements with index>=new_length
  void truncate(const int new_length) {
    if (new_length<_high && new_length>=0)
      _high = new_length;
  }

  /// add an element at the high end of the array
  void add_high(const Type& el) {
    if (full()) grow();
    _data[_high++] = el;
  }

  /**
    insert an element at the index, shifting all elements after the 
    index to high ends
   */
  void insert(const Type& el, int index) {
    if (index<0 || index>length())
      out_of_bounds(index);
    else 
      get_vacant(index) = el;
  }

  /// remove the last element 
  void del_high() { if (!empty()) _high--;}

  /// remove the element at index
  void del(int index) {
    if (index>=0 && index<length()) {
      for (Type* p = _data+index; p < _data+(_high-1); p++) 
        *p = *(p+1);
      _high--;
    }
  }
  /// Sort the elements in the array in ascending order. 
  void qsort(int (*cmp)(Type*, Type*)) {
    ::qsort(_data, _high, sizeof(Type), (int (*)(const void*, const void*)) cmp);
  }
  int bi_search(Type& key, int& index, int (*pred)(Type*, Type*))  {
    return bi_search_aux(0, length()-1, &key, index, pred);
  }

  TArray<Type>& operator= (const TArray<Type>& a)  {
    clear(a.length());
    for (int i = 0; i<a.length(); i++){
      add_high(a[i]);
    }
    return *this;
  }

  
};

/**
   An iterator for enumerating elements in a TArray. With this class
   and the 'forall' macro in "forall.h", the elements in an array can
   be enumerated in the following way:   
     forall (Type, variable, TArray, array) {
       process(variable);
     }
*/
template<class CLASS_DECLSPEC Type> class TArrayIterator : public Iterator {
protected:
  TArray<Type>* _array;
  int _current;
  Type _fake_value;
public:
  Type& current() {
    if (_array==0 || _current<0 || _current >=_array->length())
      return _fake_value;
    else
      return (*_array)[_current];
  }
  Type& del_current() {
    Type& val = current();
    if (!finished()) {
      _array->del(_current);
      updated();
    }
    return val;
  }
  int finished() const  { return _array==0 || _current>=_array->length();}
  Type& next() {
    if (++_current<_array->length()) 
      return (*_array)[_current];
    else 
      return _fake_value;
  }    
  void reset()  { _current = 0; updated(0);}
  TArrayIterator(const TArray<Type>& array)  { 
    _array = (TArray<Type>*) &array; 
    reset();
    memset(&_fake_value, 0, sizeof(Type));
  }
  TArrayIterator(const TArray<Type>* array)  { 
    if (array==0) {
      _array = 0;
      _current = -1;
    }
    else {
      _array = (TArray<Type>*) array; 
      reset();
    }
    memset(&_fake_value, 0, sizeof(Type));
  }
};

#define for2arrays(Type, x1, x2, control, array1, array2) \
Type x1=0; \
Type x2=0; \
int control=0, name2(x1,_index)=0, name2(x2,_index)=0;\
for (x1=(array1.length()>0)?&array1[0]:0,\
     x2=(array2.length()>0)?&array2[0]:0;\
name2(x1,_index)<array1.length() || name2(x2,_index)<array2.length(); \
x1=((control<=0)?((++name2(x1,_index)<array1.length())?(&array1[name2(x1,_index)]):0):x1),\
x2=((control>=0)?((++name2(x2,_index)<array2.length())?(&array2[name2(x2,_index)]):0):x2))

#define for_each_in_array (Type, x, array) \
Type x=0; int name2(x,_index)=0;\
for (x=(array.length()>0)?&array[0]:0; name2(x,_index)<array.length(); \
x=(++name2(x,_index)<array.length())?(&array[name2(x,_index)]):0)

#endif

