/************************************************************************/
/*  Copyright (c) 1997-2000 Nalante, Inc., 245 Falconer End, Edmonton,  */
/*  Alberta, Canada, T6R 2V6, Phone: (780) 432-5288, Fax: (780) 988-0184*/
/*  All Rights Reserved                                                 */
/*                                                                      */
/*  THIS IS UNPUBLISHED PROPRIETARY SOURCE CODE OF Nalante, Inc.        */
/*  The copyright notice above does not evidence any                    */
/*  actual or intended publication of such source code.                 */
/*                                                                      */
/*  Unauthorized copy or distribution of this file is prohibited.       */
/************************************************************************/



#ifndef TSLIST_H
#define TSLIST_H

#ifndef SLIST_H
#include "slist.h"
#endif

/** A TSlist is a circular single linked list. Elements can be added
    and removed at the both ends of the list efficiently. The elements
    in the list must all be pointers (i.e., Type must be some kind of
    pointer) */

template<class CLASS_DECLSPEC Type> class TSlist : public Slist {
public:
  /** Assuming the current elements in the list are in ascending
      order, a new element is inserted into the list while
      maintaining the order. If no_dup is not 0, the new element is
      not inserted into the list if there is an identical element in
      the list already. */
  Type ord_insert(const Type element, int (*cmp)(Type,Type), int no_dup = 0) {
    return (Type) Slist::ord_insert((Ent) element, (TwoArgPred) cmp, no_dup);
  }

  /** Assuming the current elements in the list are in ascending
      order, the list is searched for the element. Upon returning from
      this function, the value of place is the location where the
      element is found or where it should be inserted. The return
      value of the function is an element in the list that is
      identical to 'element' or 0 if such identical element is not
      found. */
  Type ord_search(const Type element, int (*cmp)(Type, Type), Slink*& place) {
    return (Type) Slist::ord_search((Ent) element, (TwoArgPred) cmp, place);
  }

  /** Deleting an element from an ordered TSlist */     
  Type ord_del(const Type a, int (*cmp)(Type, Type)) {
    return (Type) Slist::ord_del((Ent) a, (TwoArgPred) cmp);
  }

  /// Sorts the elements in the list 
  void sort(int (*cmp)(Type, Type)) { Slist::sort((TwoArgPred) cmp);}

  /** Insert an element at a place in the list. The value of 'place'
      should be obtained from ord_search function */
  void insert_at(Type a, Slink* place) { Slist::insert_at((Ent) a, place);}

  /// insert an element at the beginning of the list
  void insert(const Type a) { Slist::insert((Ent) a);}

  /// append the element to the end of the list
  void append(const Type a) { Slist::append((Ent) a);}

  /// insert an element at the position indicated by index.
  void insert(int index, const Type a) { Slist::insert(index, (Ent) a);}

  /// add the element to the end of the list (same as append).
  void add(const Type a) { Slist::add((Ent) a);}
  
  /** Append the list to the end of this list. Upon returning of this
      function the list will be empty. */
  void append_list(TSlist& list)  { Slist::append_list((Slist&) list);}

  /** Prepend the list to the beginning of this list. Upon returning
      of this function the list will be empty. */
  void prepend_list(TSlist& list) { Slist::prepend_list((Slist&) list);}

  /// Remove the first element from the list and return its value
  Type get() { return (Type) Slist::get();}

  /** Return the value of index'th element in the list (without
      removing it from the list).*/
  Type get(int index) const { return (Type) Slist::get(index);}

  /// Replace the index'th element in the list with a
  Type replace(int index, const Type a) {
    return (Type) Slist::replace(index, (Ent) a);
  }

  /// Delete the element from the list. 
  Type del(const Type a) { return (Type) Slist::del((Ent) a);}

  /// Search the list for a, return a if it is found, otherwise return 0
  Type search(const Type a) const { return (Type) Slist::search((Ent) a);}

  /** Search the list for an element that satisfies pred, return the
      element if it is found, otherwise return 0. If neg is not 0, the
      predicate pred is negated */
  Type search(int (*pred)(Type), int neg = 0) const {
    return (Type) Slist::search((OneArgPred) pred, neg);
  }
  
  /** Search the list for an element such that pred(element, a) is
      true, return the element if it is found, otherwise return 0. If
      neg is not 0, the predicate pred is negated */
  Type search(const Type a, int (*pred)(Type, Type), int neg = 0) const {
    return (Type) Slist::search((Ent) a, (TwoArgPred) pred, neg);
  }

  /// return the smallest element in the list (or largest if neg is not 0).
  Type find_minimum(int (*cmp)(Type, Type), int neg = 0) const {
    return (Type) Slist::find_minimum((TwoArgPred) cmp, neg);
  }

  /// empty the list.
  void clear() { Slist::clear();}

  /// return true if the list is empty
  int empty() const { return Slist::empty();}

  /// return the number of elements in the list
  int length() const { return Slist::length();}

  /// return the first element in the list
  Type first() const { return (Type) Slist::first();}

  /// return the last element in the list
  Type last() const  { return (Type) Slist::last();}

  /// make a copy of the list
  void copy(const TSlist& list) { Slist::copy((Slist&) list);}

  /// Construct an empty list
  TSlist() {}

  /// Construct a copy of list
  TSlist(const TSlist& list) : Slist((Slist&) list) {}

  /// Make a (shallow) copy of the list
  TSlist& operator = (const TSlist& list) { copy(list); return *this;}

  /// Same as get
  Type pop() { return (Type) Slist::pop();}

  /// Same as insert
  void push(Type p) { Slist::push((Ent) p);}

  /// Same as get(0)
  Type top() const { return (Type) Slist::top();}

  /** Apply the function func to the elements in the list in the
      following way:
        v[0] = seed
        v[i+1] = func(v[i], element[i])
      Return the last v value.

      For example, to compute the sum of all the elements in the list,
      the seed is 0, the func returns the sum of two elements.
  */
  Type reduce(const Type seed, Type (*func)(Type, Type)) {
    return (Type) Slist::reduce((Ent) seed, (TwoArgFunc) func);
  }

  /// reverse the order of elements in the list.
  void reverse() { Slist::reverse(); }

  void map(TSlist& a, Type (*f)(Type)) { 
    Slist::map((Slist&) a, (OneArgFunc) f);
  }

  void combine(TSlist& a, TSlist& b, int (*f)(Type, Type)) {
    Slist::combine((Slist&) a, (Slist&) b, (TwoArgFunc) f);
  }

  void remove(Type arg, int (*pred)(Type, Type), TSlist& removed) {
    Slist::remove((Ent) arg, (TwoArgPred) pred, (Slist&) removed);
  }

  void select(TSlist& b,  int (*pred)(Type), int neg = 0) {
    Slist::select((Slist&) b, (OneArgPred) pred, neg);
  }

  void select(TSlist& b, Type arg, int (*pred)(Type, Type), int neg = 0) {
    Slist::select((Slist&) b, (Ent) arg, (TwoArgPred) pred, neg);
  }

  void collect(TSlist& b, Type (*func)(Type)) {
    Slist::collect((Slist&) b, (OneArgFunc) func);
  }

  void collect(TSlist& b, Type arg, Type (*func)(Type, Type)) {
    Slist::collect((Slist&) b, (Ent) arg, (TwoArgFunc) func);
  }

  void apply(void (*proc)(Type)) { Slist::apply((OneArgProc) proc); }

  void apply(Type arg, void (*proc)(Type, Type)) {
    Slist::apply((Ent) arg, (TwoArgProc) proc);
  }

  void merge(TSlist& a, TSlist& b, int (*cmp)(Type, Type)) {
    Slist::merge((Slist&) a, (Slist&) b, (TwoArgPred) cmp);
  }

};

template<class CLASS_DECLSPEC Type> class TSlistIterator : public SlistIterator {
public:
  TSlistIterator<Type>(const TSlist<Type>* s) : SlistIterator((Slist*) s) {};
  TSlistIterator<Type>(const TSlist<Type>& s) : SlistIterator((Slist&) s) {};
  Type next() { return (Type) SlistIterator::next();}
  Type current() const { return (Type) SlistIterator::current();}
  void reset(const TSlist<Type>* s = 0) { SlistIterator::reset((Slist*) s);}
  void reset(const TSlist<Type>& s) { reset(&s);}
  Type del_current() { return (Type) SlistIterator::del_current();}
  void insert_before_current(Type x) {
    SlistIterator::insert_before_current((Ent) x);
  }
  void insert_after_current(Type x) {
    SlistIterator::insert_after_current((Ent) x);
  }
};

template<class CLASS_DECLSPEC Type> class TSlistFollowingIterator : public SlistFollowingIterator {
public:
  TSlistFollowingIterator<Type>(TSlistIterator<Type>& s) : SlistFollowingIterator(s) {};
  Type next() { return (Type) SlistIterator::next();}
  Type current() const { return (Type) SlistIterator::current();}
  Type del_current() { return (Type) SlistFollowingIterator::del_current();}
};

#endif /* TSLIST_H */
